from __future__ import annotations

import ast
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from thriftpyi.proxies import TModuleProxy


def build(
    proxy: TModuleProxy, is_async: bool, strict_fields: bool, strict_methods: bool
) -> ast.Module:
    return ast.Module(
        body=[
            *_make_imports(proxy),
            *_make_exceptions(proxy, strict=strict_fields),
            *_make_enums(proxy),
            *_make_structs(proxy, strict=strict_fields),
            *_make_consts(proxy),
            *_make_service(proxy, is_async, strict=strict_methods),
        ],
        type_ignores=[],
    )


def build_init(imports: Iterable[str]) -> ast.Module:
    return ast.Module(
        body=[_make_relative_import(imports)],
        type_ignores=[],
    )


def build_typedefs() -> ast.Module:
    types_map = {
        "Binary": ("bytes", "BINARY"),
        "Bool": ("bool", "BOOL"),
        "Byte": ("int", "BYTE"),
        "Double": ("float", "DOUBLE"),
        "I16": ("int", "I16"),
        "I32": ("int", "I32"),
        "I64": ("int", "I64"),
        "String": ("str", "STRING"),
    }

    body: list[ast.stmt] = [
        ast.ImportFrom(
            module="dataclasses",
            names=[ast.alias(name="dataclass")],
            level=0,
        ),
        ast.ImportFrom(
            module="typing",
            names=[ast.alias(name="Annotated")],
            level=0,
        ),
        ast.ClassDef(
            name="Metadata",
            bases=[],
            keywords=[],
            body=[
                ast.AnnAssign(
                    target=ast.Name(id="thrift_type"),
                    annotation=ast.Name(id="str"),
                    simple=1,
                ),
            ],
            decorator_list=[
                ast.Call(
                    func=ast.Name(id="dataclass"),
                    args=[],
                    keywords=[
                        ast.keyword(
                            arg="frozen",
                            value=ast.Constant(value=True),
                        ),
                    ],
                ),
            ],
        ),
    ]

    for alias, (base_type, thrift_type) in types_map.items():
        assign = ast.Assign(
            targets=[ast.Name(id=alias)],
            value=ast.Subscript(
                value=ast.Name(id="Annotated"),
                slice=ast.Tuple(
                    elts=[
                        ast.Name(id=base_type),
                        ast.Call(
                            func=ast.Name(id="Metadata"),
                            args=[],
                            keywords=[
                                ast.keyword(
                                    arg="thrift_type",
                                    value=ast.Constant(value=thrift_type),
                                ),
                            ],
                        ),
                    ],
                ),
            ),
            lineno=0,
        )
        body.append(assign)

    return ast.Module(body=body, type_ignores=[])


def _make_imports(proxy: TModuleProxy) -> list[ast.ImportFrom]:
    imports = []
    if proxy.has_structs():
        imports.append(_make_absolute_import("dataclasses", "dataclass, field"))
    if proxy.has_enums():
        imports.append(_make_absolute_import("enum", "IntEnum"))

    imports.append(_make_absolute_import("typing", "*"))
    imports.append(_make_relative_import(["_typedefs"]))
    imports.extend(
        _make_relative_import([name]) for name in sorted(proxy.get_imports())
    )

    return imports


def _make_absolute_import(module: str, name: str) -> ast.ImportFrom:
    return ast.ImportFrom(
        module=module,
        names=[ast.alias(name=name, asname=None)],
        level=0,
    )


def _make_relative_import(names: Iterable[str]) -> ast.ImportFrom:
    return ast.ImportFrom(
        module=None,
        names=[ast.alias(name=name, asname=None) for name in sorted(names)],
        level=1,
    )


def _make_consts(interface: TModuleProxy) -> list[ast.stmt]:
    return [item.as_ast() for item in interface.get_consts()]


def _make_exceptions(interface: TModuleProxy, strict: bool) -> list[ast.ClassDef]:
    return [
        item.with_options(ignore_required=not strict).as_ast(bases=["Exception"])
        for item in interface.get_exceptions()
    ]


def _make_enums(interface: TModuleProxy) -> list[ast.ClassDef]:
    return [item.as_ast(bases=["IntEnum"]) for item in interface.get_enums()]


def _make_structs(interface: TModuleProxy, strict: bool) -> list[ast.ClassDef]:
    return [
        item.with_options(ignore_required=not strict).as_ast(decorators=["dataclass"])
        for item in interface.get_structs()
    ]


def _make_service(
    interface: TModuleProxy, is_async: bool, strict: bool
) -> list[ast.ClassDef]:
    services = interface.get_services()

    return [
        item.with_options(ignore_optional=strict, is_async=is_async).as_ast()
        for item in services
    ]



from __future__ import annotations

import ast
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Hashable, Sequence, Type, Union

if TYPE_CHECKING:
    AnyFunctionDef = Union[ast.AsyncFunctionDef, ast.FunctionDef]
    AnyFunctionDefType = Union[Type[ast.AsyncFunctionDef], Type[ast.FunctionDef]]


FieldValue = Union[str, int, bool, None]


@dataclass
class ModuleItem:
    name: str
    fields: list[Field] = field(default_factory=list)
    methods: list[Method] = field(default_factory=list)

    def with_options(
        self,
        *,
        ignore_optional: bool = False,
        ignore_required: bool = False,
        is_async: bool = False,
    ) -> ModuleItem:
        return self.__class__(
            name=self.name,
            fields=[
                field.with_options(
                    ignore_optional=ignore_optional,
                    ignore_required=ignore_required,
                )
                for field in self.fields
            ],
            methods=[
                method.with_options(
                    ignore_optional=ignore_optional,
                    ignore_required=ignore_required,
                    is_async=is_async,
                )
                for method in self.methods
            ],
        )

    def as_ast(
        self,
        *,
        bases: Sequence[str] = None,
        decorators: Sequence[str] = None,
    ) -> ast.ClassDef:
        bases = bases or []
        decorators = decorators or []

        body: list[ast.stmt] = []
        body.extend(entry.as_ast() for entry in self.fields)
        body.extend(entry.as_ast() for entry in self.methods)

        if not body:
            body = [ast.Expr(value=ast.Constant(value=Ellipsis))]

        return ast.ClassDef(
            name=self.name,
            bases=[ast.Name(id=base, ctx=ast.Load()) for base in bases],
            keywords=[],
            body=body,
            decorator_list=[
                ast.Name(id=decorator, ctx=ast.Load()) for decorator in decorators
            ],
        )


@dataclass
class Method:
    name: str
    args: list[Field] = field(default_factory=list)
    returns: list[Field] = field(default_factory=list)
    is_async: bool = False

    def with_options(
        self,
        *,
        ignore_optional: bool = False,
        ignore_required: bool = False,
        is_async: bool = False,
    ) -> Method:
        return self.__class__(
            name=self.name,
            args=[
                entry.with_options(
                    ignore_optional=ignore_optional,
                    ignore_required=ignore_required,
                )
                for entry in self.args
            ],
            returns=[
                entry.with_options(
                    ignore_optional=ignore_optional,
                    ignore_required=ignore_required,
                )
                for entry in self.returns
            ],
            is_async=is_async,
        )

    def as_ast(self) -> AnyFunctionDef:
        factory: AnyFunctionDefType = ast.FunctionDef
        if self.is_async:
            factory = ast.AsyncFunctionDef

        returns: ast.expr
        if self.returns and self.returns[0].type:
            returns = ast.Name(id=self.returns[0].type, ctx=ast.Load())
        else:
            returns = ast.Constant(value=None, kind=None)

        return factory(
            name=self.name,
            args=ast.arguments(  # type: ignore[call-overload]
                posonlyargs=[],
                args=[
                    ast.arg(arg="self", annotation=None),
                    *[entry.as_ast() for entry in self.args],
                ],
                vararg=[],
                kwarg=[],
                kwonlyargs=[],
                kw_defaults=[],
                defaults=[],
            ),
            body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
            decorator_list=[],
            returns=returns,
            lineno=0,
        )


@dataclass
class Field:
    name: str
    type: str | None
    value: FieldValue
    required: bool
    module: str | None = None

    def with_options(self, *, ignore_optional: bool, ignore_required: bool) -> Field:
        required = self.required
        if ignore_optional and not required:
            required = True
        if ignore_required and required:
            required = False

        return self.__class__(
            name=self.name,
            type=self.type,
            value=self.value,
            required=required,
            module=self.module,
        )

    def as_ast(self) -> ast.AnnAssign | ast.Assign:
        if self.type is None:
            return ast.Assign(
                targets=[ast.Name(id=self.name, ctx=ast.Store())],
                value=self._make_ast_value(),
                lineno=0,
            )

        if not self.required:
            annotation = ast.Name(id=f"Optional[{self.type}]", ctx=ast.Load())
        else:
            annotation = ast.Name(id=self.type, ctx=ast.Load())

        if self.required and self.value is None:
            value = None
        else:
            value = self._make_ast_value()

        return ast.AnnAssign(
            target=ast.Name(id=self.name, ctx=ast.Store()),
            annotation=annotation,
            value=value,
            simple=1,
        )

    def _make_ast_value(self) -> ast.expr:
        if self.module:
            body = ast.parse(f"{self.module}.{self.value}").body
            return body[0].value  # type: ignore[attr-defined, no-any-return]
        return ast.Constant(value=self.value, kind=None)


@dataclass
class StructField(Field):
    def _make_ast_value(self) -> ast.expr:
        if not isinstance(self.value, Hashable):
            if self.value:
                value = ast.Lambda(
                    args=[], body=ast.Constant(value=self.value, kind=None)
                )
            else:
                value = ast.Name(id=self.value.__class__.__name__, ctx=ast.Load())

            return ast.Call(
                func=ast.Name(id="field", ctx=ast.Load()),
                args=[],
                keywords=[
                    ast.keyword(
                        arg="default_factory",
                        value=value,
                    )
                ],
            )
        return super()._make_ast_value()



from __future__ import annotations

from pathlib import Path


def list_interfaces(interfaces_dir: str | Path) -> list[Path]:
    return list(Path(interfaces_dir).glob("**/*.thrift"))


def save(data: str, to: str | Path) -> None:
    Path.mkdir(Path(to).parent, exist_ok=True)
    with open(to, "w+", encoding="utf-8") as f:
        f.write(data)



from __future__ import annotations

from typing import cast

from thriftpyi.entities import Field, FieldValue, Method, ModuleItem, StructField
from thriftpyi.utils import get_python_type, guess_type


class TModuleProxy:
    __slots__ = ["tmodule"]

    def __init__(self, tmodule) -> None:
        self.tmodule = tmodule

    def get_consts(self) -> list[Field]:
        tconsts = (
            (name, value)
            for name, value in vars(self.tmodule).items()
            if value in self.tmodule.__thrift_meta__["consts"]
        )
        return [self._make_const(tconst) for tconst in tconsts]

    def get_enums(self) -> list[ModuleItem]:
        return [
            self._make_enum(tenum) for tenum in self.tmodule.__thrift_meta__["enums"]
        ]

    def get_exceptions(self) -> list[ModuleItem]:
        return [
            self._make_exception(texc)
            for texc in self.tmodule.__thrift_meta__["exceptions"]
        ]

    def get_imports(self) -> list[str]:
        return [item.__name__ for item in self.tmodule.__thrift_meta__["includes"]]

    def get_services(self) -> list[ModuleItem]:
        return [
            self._make_service(tservice)
            for tservice in self.tmodule.__thrift_meta__["services"]
        ]

    def get_structs(self) -> list[ModuleItem]:
        return [
            self._make_struct(tstruct)
            for tstruct in self.tmodule.__thrift_meta__["structs"]
        ]

    def has_structs(self) -> bool:
        return len(self.tmodule.__thrift_meta__["structs"]) > 0

    def has_enums(self) -> bool:
        return len(self.tmodule.__thrift_meta__["enums"]) > 0

    def _make_const(self, tconst) -> Field:
        name, value = tconst

        known_modules = {
            module.__name__ for module in self.tmodule.__thrift_meta__["includes"]
        }

        module_name = None
        if hasattr(value, "__class__"):
            module_name = value.__class__.__module__
            if module_name not in known_modules:
                module_name = None

        return Field(
            name=name,
            type=guess_type(
                value,
                known_modules=known_modules,
                known_structs=self.tmodule.__thrift_meta__["structs"],
            ),
            value=value,
            module=module_name,
            required=True,
        )

    @staticmethod
    def _make_enum(tenum) -> ModuleItem:
        fields = tenum._NAMES_TO_VALUES  # pylint: disable=protected-access
        ttype = tenum._ttype  # pylint: disable=protected-access
        tspec = {idx: (ttype, name, True) for idx, name in enumerate(fields)}
        spec = TSpecProxy(
            module_name=tenum.__module__,
            thrift_spec=tspec,
            default_spec=fields,
        )
        return ModuleItem(name=tenum.__name__, fields=spec.get_fields(ignore_type=True))

    @staticmethod
    def _make_exception(texc) -> ModuleItem:
        spec = TSpecProxy(
            module_name=texc.__module__,
            thrift_spec=texc.thrift_spec,
            default_spec=dict(texc.default_spec),
        )

        fields = spec.get_fields()
        methods = []
        if fields:
            methods.append(Method(name="__init__", args=fields))

        return ModuleItem(name=texc.__name__, methods=methods, fields=fields)

    @staticmethod
    def _make_service(tservice) -> ModuleItem:
        return ModuleItem(
            name=tservice.__name__,
            methods=[
                Method(
                    name=method_name,
                    args=TSpecProxy(
                        module_name=tservice.__module__,
                        thrift_spec=getattr(
                            tservice, f"{method_name}_args"
                        ).thrift_spec,
                        default_spec=dict(
                            getattr(tservice, f"{method_name}_args").default_spec
                        ),
                    ).get_fields(),
                    returns=TSpecProxy(
                        module_name=tservice.__module__,
                        thrift_spec=getattr(
                            tservice, f"{method_name}_result"
                        ).thrift_spec,
                        default_spec=dict(
                            getattr(tservice, f"{method_name}_args").default_spec
                        ),
                    ).get_fields(),
                )
                for method_name in tservice.thrift_services
            ],
        )

    @staticmethod
    def _make_struct(tclass) -> ModuleItem:
        spec = TStructSpecProxy(
            module_name=tclass.__module__,
            thrift_spec=tclass.thrift_spec,
            default_spec=dict(tclass.default_spec),
        )
        return ModuleItem(
            name=tclass.__name__,
            fields=spec.get_fields(),
        )


class TSpecItemProxy:
    __slots__ = ("ttype", "name", "meta", "required")

    def __init__(self, item: tuple):
        ttype, name, *meta, required = item
        self.ttype = ttype
        self.name = name
        self.meta = meta
        self.required = required


class TSpecProxy:
    __slots__ = ("module_name", "thrift_spec", "default_spec")

    def __init__(self, module_name: str, thrift_spec, default_spec):
        self.module_name = module_name
        self.thrift_spec = [TSpecItemProxy(thrift_spec[k]) for k in sorted(thrift_spec)]
        self.default_spec = default_spec

    def get_fields(self, *, ignore_type: bool = False) -> list[Field]:
        return [
            Field(
                name=item.name,
                type=self._get_python_type(item) if not ignore_type else None,
                value=self._get_default_value(item),
                required=item.required,
            )
            for item in self.thrift_spec
        ]

    def _remove_self_module(self, pytype: str) -> str:
        left_type, sep, right_type = pytype.partition(",")
        # Due to complex type, such as Dict[some_module.TypeA, some_module.TypeB]
        # recursively deal with the first and second parts
        if right_type != "":
            return (
                self._remove_self_module(left_type)
                + sep
                + self._remove_self_module(right_type)
            )
        start, _, end = pytype.rpartition(f"{self.module_name}.")
        return start + end

    def _get_python_type(self, item: TSpecItemProxy) -> str:
        pytype = get_python_type(item.ttype, meta=item.meta)
        return self._remove_self_module(pytype)

    def _get_default_value(self, item: TSpecItemProxy) -> FieldValue:
        default_value = self.default_spec.get(item.name)
        return cast(FieldValue, default_value)


class TStructSpecProxy(TSpecProxy):
    def get_fields(self, *, ignore_type: bool = False) -> list[Field]:
        return [
            StructField(
                name=item.name,
                type=self._get_python_type(item) if not ignore_type else None,
                value=self._get_default_value(item),
                required=item.required,
            )
            for item in self.thrift_spec
        ]



from collections.abc import Mapping
from typing import Any, Collection, List, Tuple, Type

from thriftpy2.thrift import TType


def guess_type(  # pylint: disable=too-many-branches
    value, *, known_modules: Collection[str], known_structs: Collection[Type[Any]]
) -> str:
    if isinstance(value, (bool, int, float, str, bytes)):
        return type(value).__name__

    if isinstance(value, Mapping):
        type_ = type(value).__name__.capitalize()
        key_type = guess_type(
            next(iter(value.keys())),
            known_modules=known_modules,
            known_structs=known_structs,
        )
        value_type = guess_type(
            next(iter(value.values())),
            known_modules=known_modules,
            known_structs=known_structs,
        )
        return f"{type_}[{key_type}, {value_type}]"

    if isinstance(value, Collection):
        type_ = type(value).__name__.capitalize()
        item_type = guess_type(
            next(iter(value)), known_modules=known_modules, known_structs=known_structs
        )
        return f"{type_}[{item_type}]"

    if hasattr(value, "__class__"):
        module_name: str = value.__class__.__module__
        class_name: str = value.__class__.__name__
        if module_name in known_modules:
            return f"{module_name}.{class_name}"
        if type(value) in known_structs:
            return class_name
    return "Any"


def get_python_type(ttype: int, meta: List) -> str:
    return TTYPE_MAP[ttype](meta)


def _get_bool(meta: List) -> str:
    del meta
    return "_typedefs.Bool"


def _get_double(meta: List) -> str:
    del meta
    return "_typedefs.Double"


def _get_byte(meta: List) -> str:
    del meta
    return "_typedefs.Byte"


def _get_binary(meta: List) -> str:
    del meta
    return "_typedefs.Binary"


def _get_i16(meta: List) -> str:
    del meta
    return "_typedefs.I16"


def _get_i32(meta: List) -> str:
    if meta and meta[0] is not None:
        return f"{meta[0].__module__}.{meta[0].__name__}"
    return "_typedefs.I32"


def _get_i64(meta: List) -> str:
    del meta
    return "_typedefs.I64"


def _get_str(meta: List) -> str:
    del meta
    return "_typedefs.String"


def _get_struct(meta: List) -> str:
    return f"{meta[0].__module__}.{meta[0].__name__}"


def _get_list(meta: List) -> str:
    subttype, submeta = _unpack_meta(meta)
    return f"List[{get_python_type(subttype, submeta)}]"


def _get_map(meta: List) -> str:
    key, value = meta[0]
    key_ttype, key_meta = _unpack_meta([key])
    value_ttype, value_meta = _unpack_meta([value])
    key_pytype = get_python_type(key_ttype, key_meta)
    value_pytype = get_python_type(value_ttype, value_meta)
    return f"Dict[{key_pytype}, {value_pytype}]"


def _get_set(meta: List) -> str:
    subttype, submeta = _unpack_meta(meta)
    return f"Set[{get_python_type(subttype, submeta)}]"


def _unpack_meta(meta: List) -> Tuple[int, List]:
    try:
        subttype, submeta = meta[0]
    except TypeError:
        subttype, submeta = meta[0], None
    return subttype, [submeta]


def _register_binary(mapping):
    ttype_binary = getattr(TType, "BINARY", TType.STRING)
    if ttype_binary != TType.STRING:
        mapping[ttype_binary] = _get_binary


TTYPE_MAP = {
    TType.BOOL: _get_bool,
    TType.DOUBLE: _get_double,
    TType.BYTE: _get_byte,
    TType.I16: _get_i16,
    TType.I32: _get_i32,
    TType.I64: _get_i64,
    TType.STRING: _get_str,
    TType.STRUCT: _get_struct,
    TType.MAP: _get_map,
    TType.SET: _get_set,
    TType.LIST: _get_list,
}


_register_binary(TTYPE_MAP)



from __future__ import annotations

import ast
import subprocess
import sys
from pathlib import Path

import thriftpy2

from thriftpyi import files, stubs
from thriftpyi.proxies import TModuleProxy


def thriftpyi(  # pylint: disable=too-many-locals
    interfaces_dir: str,
    output_dir: Path,
    is_async: bool = False,
    strict_fields: bool = False,
    strict_methods: bool = True,
) -> None:
    output_dir = output_dir.resolve()
    interfaces = files.list_interfaces(interfaces_dir)

    stub = stubs.build_init(path.stem for path in interfaces)
    files.save(ast.unparse(stub), to=output_dir / "__init__.pyi")

    stub = stubs.build_typedefs()
    files.save(ast.unparse(stub), to=output_dir / "_typedefs.pyi")

    for path in interfaces:
        tmodule = thriftpy2.load(str(path))
        proxy = TModuleProxy(tmodule)
        stub = stubs.build(
            proxy,
            is_async=is_async,
            strict_fields=strict_fields,
            strict_methods=strict_methods,
        )
        files.save(ast.unparse(stub), to=output_dir / path.with_suffix(".pyi").name)

    lint(output_dir)


def lint(output_dir: Path) -> None:
    python_executable = sys.executable
    if not python_executable:
        raise RuntimeError(
            "Python executable not found. Make sure that Python is "
            "installed and the path is set correctly."
        )

    subprocess.check_call(
        [f"{python_executable} -m autoflake -i -r {output_dir.joinpath('*')}"],
        shell=True,
    )
    subprocess.check_call(
        [
            f"{python_executable}",
            "-m",
            "black",
            "--pyi",
            "--quiet",
            *list(output_dir.glob("*.pyi")),
        ]
    )



from __future__ import annotations

import argparse
from pathlib import Path

from thriftpyi.main import thriftpyi


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("interfaces_dir")
    parser.add_argument(
        "-o",
        "--output",
        dest="output_dir",
        type=Path,
        help="Directory where to save generated `.pyi` files",
    )
    parser.add_argument(
        "--async",
        dest="is_async",
        action="store_true",
        help="Whether service methods should be async or not",
    )
    parser.add_argument(
        "--strict-optional",
        dest="strict_optional",
        action="store_true",
        help="If not specified all fields will be optional with default to None",
    )
    return parser


def main(argv=None) -> int:
    parser = make_parser()
    args = parser.parse_args(argv)
    thriftpyi(
        args.interfaces_dir,
        args.output_dir,
        is_async=args.is_async,
        strict_fields=args.strict_optional,
        strict_methods=True,
    )
    return 0
